----------------------------------------------------------------------------
-- Barret Reducer (barret_reducer.vhd)
--
-- Computes the remainder (x mod m) using the 
-- Barret algorithm (combinatorial implementation)

-- As C and M are constants, 
-- more efficient multipliers can be implemented.
----------------------------------------------------------------------------



library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.gcd_pkg.all;


entity Barrett_reducer_comb is
port (
  x: in std_logic_vector (N-1 downto 0);
  z: out std_logic_vector (K-1 downto 0)
);
end Barrett_reducer_comb;

architecture rtl of Barrett_reducer_comb is
  signal mult_1: std_logic_vector(2*(N-K) + 2 downto 0);
  signal mult_2: std_logic_vector(2*K + 1 downto 0);
  signal q: std_logic_vector(K+1 downto 0);
  signal r, product: std_logic_vector(K+1 downto 0);
  signal r_minus_m, r_minus_2m: std_logic_vector(K+2 downto 0);
   
--  signal C: std_logic_vector(N-K downto 0) :=  (others => '0');
  --For Xilinx XST synthesis
  attribute mult_style: string;
  attribute mult_style of mult_1: signal is "kcm"; --"{auto|block|pipe_block|kcm|csd|lut|pipe_lut}";
  attribute mult_style of mult_2: signal is "csd"; 

begin
 
  mult_1 <= x(N-1 downto K-1) * ('0' & C);
  q <= mult_1(N+2 downto N-K+1);
  
  mult_2 <= q * M;

  product <= mult_2(K+1 downto 0);

  r <= x(k+1 downto 0) - product;

  r_minus_m <= ('0' & r) - M;
  r_minus_2m <= ('0' & r) - (M & '0');

  muxout: process(r, r_minus_m, r_minus_2m)
  begin
  if r_minus_m(K+2) = '1' then 
    Z <= r(K-1 downto 0);
  elsif r_minus_2m(K+2) = '1' then
    Z <= r_minus_m(K-1 downto 0);
  else
    Z <= r_minus_2m(K-1 downto 0);
  end if;
  end process;

end rtl;

