----------------------------------------------------------------------------------
-- This file is barrelshifter  

-- Copyright (C) 2021
-- Authors: Sanjay Deshpande <sanjay.deshpande@yale.edu>
          
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software Foundation,
-- Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_unsigned.ALL;
use IEEE.NUMERIC_STD.ALL;

entity barrelshifter is
GENERIC ( W : integer := 1024;
          logW : integer := 10;
          t   : integer := 16
          );
        port(
        			shift  : in std_logic_vector(logW-1 downto 0);
        			inword  : in std_logic_vector(W-1 downto 0);
        			outword : out std_logic_vector(W-1 downto 0) 
        		);
--    attribute use_dsp : string;
--    attribute use_dsp of barrelshifter : entity is "yes";
end barrelshifter;

architecture barrel of barrelshifter is
signal shift_int : integer range 0 to t;
begin

shift_int <= to_integer(unsigned(shift));
barr: for position in 0 to inword'high generate
       outword(position) <= inword((position - shift_int) mod (inword'high + 1)) when ((position  > shift_int) or (position = shift_int)) else '0'  ;
 end generate barr;

end barrel;

