----------------------------------------------------------------------------------
-- This file is full width module for divstep2 function from fast GCD inverse for integers algorithm 

-- Copyright (C) 2021
-- Authors: Sanjay Deshpande <sanjay.deshpande@yale.edu>
          
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software Foundation,
-- Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

-- K  =  input width
-- t  = value from iterations function of fast gcd algorithm
-- logt = log2(t)
-- K_adder_slices = number of slices K should be divided in
-- t_adder_slices = number of slices t should be divided in
-- K_adder_width  =  matching the adder width using following formula (K + (K_adder_slices-(K mod K_adder_slices)))
-- t_adder_width  =  matching the adder width using following formula (t + (t_adder_slices-(t mod t_adder_slices)))
-- barrel_shift_width = variable shifter width 
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

entity divsteps2 is
    generic ( 
              --           
              K : integer := 255; 
              t : integer := 738; 
              logt : integer := 10; 
              K_adder_slices : integer := 3; 
              t_adder_slices : integer := 3;
              K_adder_width : integer := 258; 
              t_adder_width : integer := 741; 
              barrel_shift_width  : integer := 10 
              
--              K : integer := 434;
--              logt : integer := 11;
--              t : integer := 1254;
--              K_adder_width : integer := 435;
--              t_adder_width : integer := 1255;
--              K_adder_slices : integer := 5;
--              t_adder_slices : integer := 5;
--              barrel_shift_width  : integer := 11 
              
    );    
    Port ( clk          : in STD_LOGIC;
           rst          : in STD_LOGIC;
           start        : in STD_LOGIC;
           done         : out STD_LOGIC;
           n_in         : in STD_LOGIC_VECTOR (logt-1 downto 0);
           t_in         : in STD_LOGIC_VECTOR (logt-1 downto 0);
           delta_in     : in STD_LOGIC_VECTOR (logt-1 downto 0);
           f_in         : in STD_LOGIC_VECTOR (K-1 downto 0);
           g_in         : in STD_LOGIC_VECTOR (K-1 downto 0);
           f_out        : out STD_LOGIC_VECTOR (K-1 downto 0);
           M2Q_v_out    : out STD_LOGIC_VECTOR (t downto 0);
           v_shift_count : out STD_LOGIC_VECTOR (logt-1 downto 0)          
           );
end divsteps2;

architecture Behavioral of divsteps2 is
signal f_in_reg_2: signed(K downto 0) := (others=> '0');
signal g_in_reg_2: signed(K downto 0) := (others=> '0');
signal n_count : std_logic_vector(logt-1 downto 0) := (others => '0');
signal t_count : std_logic_vector(logt-1 downto 0) := (others => '0');
constant one    : std_logic_vector(logt-1 downto 0) := (0 => '1', others => '0');
signal en_n, init, en_t : std_logic := '0';
signal  done_s  : std_logic := '0';
signal  load  : std_logic := '0';
type t_state is (S_RESET, S_WAIT_START, S_MINUS_INIT, S_PROCESS, S_STALL_1, S_STALL_2, S_DONE, S_WAIT_ADD);
signal state        			: t_state;
signal delta: signed(logt downto 0) := (others=> '0');
signal shifted_v: signed(t downto 0) := (others=> '0');
signal shifted_r: signed(t downto 0) := (others=> '0');
signal v: signed(t downto 0) := (others=> '0');
signal r: signed(t downto 0) := (others=> '0'); 
signal en_op_0 : std_logic;
signal en_op_1 : std_logic;
signal en_op_2 : std_logic;
--========================Adder 1 signals========================
signal start_add, done_add_1 : std_logic := '0';
signal gf_adder_out : std_logic_vector(K_adder_width-1 downto 0) := (others => '0');
signal  g_in_reg_2_to_adder : std_logic_vector(K_adder_width-1 downto 0) := (others => '0');
signal  f_in_reg_2_to_adder : std_logic_vector(K_adder_width-1 downto 0) := (others => '0');
--========================Adder 2 signals========================
signal done_add_rv : std_logic := '0';
signal rv_adder_out : std_logic_vector(t_adder_width-1 downto 0) := (others => '0');
signal r_in_adder : std_logic_vector(t_adder_width-1 downto 0) := (others => '0');
signal v_in_adder : std_logic_vector(t_adder_width-1 downto 0) := (others => '0');

signal count_v: signed(logt downto 0) := (others=> '0');
signal count_r: signed(logt downto 0) := (others=> '0');
signal count_diff_v_r: signed(logt downto 0) := (others=> '0');
signal v_dr_shift_count: unsigned(logt-1 downto 0) := (others=> '0');
signal count_diff_v_r_std : std_logic_vector(logt downto 0);
signal v_std              : std_logic_vector(2**barrel_shift_width-1 downto 0) := (others => '0');   
signal shifted_v_std      : std_logic_vector(2**barrel_shift_width-1 downto 0) := (others => '0');
signal r_std              : std_logic_vector(2**barrel_shift_width-1 downto 0) := (others => '0');   
signal shifted_r_std      : std_logic_vector(2**barrel_shift_width-1 downto 0) := (others => '0');
signal dg_flag : std_logic := '0';
signal r_dr_lt_v_dr : std_logic := '0';
signal v_or_r              : std_logic_vector(2**barrel_shift_width-1 downto 0) := (others => '0'); 
signal tempr: signed(t downto 0) := (others=> '0');
signal requal0 : std_logic := '0';
signal tempv: signed(t downto 0) := (others=> '0');
signal vequal0 : std_logic := '0';
constant zero_t : std_logic_vector(2**barrel_shift_width-1 downto 0) := (others => '0'); 
constant one_t : std_logic_vector(2**barrel_shift_width-1 downto 0) := (0 => '1', others => '0'); 
signal notf: std_logic_vector(K downto 0) := (others=> '0');
signal notv: std_logic_vector(t downto 0) := (others=> '0');
signal minus : std_logic := '0';
begin
v_shift_count <= std_logic_vector(v_dr_shift_count);
f_out <= std_logic_vector(f_in_reg_2(K-1 downto 0));
M2Q_v_out <= std_logic_vector(v);
count_diff_v_r <= count_v - count_r when count_v - count_r > 0 else count_r - count_v;

process(clk)
begin
	if rising_edge(clk) then
		count_diff_v_r_std <= std_logic_vector(count_diff_v_r(count_diff_v_r_std'high downto 0));
	end if;
end process;

v_or_r <= v_std when count_r > count_v else r_std;
r_std(t-1 downto 0) <= std_logic_vector(r(t-1 downto 0));
v_std(t-1 downto 0) <= std_logic_vector(v(t-1 downto 0));

barrel_shift_v: entity work.barrelshifter(barrel) 
        GENERIC MAP( W => 2**barrel_shift_width,
                     logW => logt,
                     t   => t
                    )
        port map(
        			shift => count_diff_v_r_std(logt-1 downto 0),
        			inword   => v_or_r,
        			outword  => shifted_v_std
        		);

shifted_v(shifted_v'high) <= v(shifted_v'high); --preserving sign        		
shifted_v(t-1 downto 0) <= signed(shifted_v_std(t-1 downto 0));

shifted_r(shifted_r'high) <= r(shifted_r'high); --preserving sign        		
shifted_r(t-1 downto 0) <= signed(shifted_v_std(t-1 downto 0));


r_std(t-1 downto 0) <= std_logic_vector(r(t-1 downto 0));
notf(K downto 0) <= not std_logic_vector(f_in_reg_2(K downto 0));					
g_in_reg_2_to_adder(K downto 0) <= one_t(K downto 0) when minus = '1' else std_logic_vector(g_in_reg_2);
f_in_reg_2_to_adder(K downto 0) <= notf(K downto 0) when minus = '1' else std_logic_vector(f_in_reg_2);
  
adder_unit_1: entity work.adder(behavioral) GENERIC MAP(w => K_adder_width, adder_width => K_adder_width/K_adder_slices)
	PORT MAP (
	  clk => clk,
	  start => start_add,
	  done => done_add_1,
	  a_in => g_in_reg_2_to_adder, 
	  b_in => f_in_reg_2_to_adder, 
	  c_out => gf_adder_out
	);

notv(t downto 0) <= not std_logic_vector(v);

v_in_adder(t downto 0) <=  notv(t downto 0) when minus = '1' else
						   std_logic_vector(shifted_v(t downto 0)) when count_r > count_v else
						   std_logic_vector(v(t downto 0));
						   
r_in_adder(t downto 0) <= one_t(t downto 0) when minus = '1' else
						  std_logic_vector(shifted_r(t downto 0)) when count_r < count_v else
						  std_logic_vector(r(t downto 0));
	                           
adder_unit_3: entity work.adder(behavioral) GENERIC MAP(w => t_adder_width, adder_width => t_adder_width/t_adder_slices)
	PORT MAP (
	  clk => clk,
	  start => start_add,
	  done => done_add_rv,
	  a_in => r_in_adder, 
	  b_in => v_in_adder, 
	  c_out => rv_adder_out
	);

dg_flag <= '1' when signed(delta) > 0 and g_in_reg_2(0) = '1' else '0';

f_register: process(clk)
			begin
				if rising_edge(clk) then
					if load = '1' then
						f_in_reg_2(K-1 downto 0) <= signed(f_in);
				
					elsif en_op_0 = '1' then
                        if dg_flag = '1' then
							f_in_reg_2 <= g_in_reg_2;
						end if;
						
					end if;
				end if;
			end process;

g_register: process(clk)
			begin
				if rising_edge(clk) then
					if load = '1' then
						g_in_reg_2(K-1 downto 0) <= signed(g_in);

					
					elsif en_op_0 = '1' then
                        if dg_flag = '1' then
                            g_in_reg_2 <= signed(gf_adder_out(K downto 0));
						end if;
					
					elsif en_op_1 ='1' then 
						if  g_in_reg_2(0) = '1' then
                                g_in_reg_2 <= signed(gf_adder_out(K downto 0));
						end if;
					
					elsif en_op_2 = '1' then
                         g_in_reg_2 <= g_in_reg_2(K) & g_in_reg_2(K downto 1);
					end if;
				end if;
			end process;

delta_register: process(clk)
			begin
				if rising_edge(clk) then
					if load = '1' then
						delta(logt-1 downto 0) <= signed(delta_in);
					
					elsif en_op_0 = '1' then
                        if dg_flag = '1' then
							delta <= -delta;
						end if;
					
					elsif en_op_1 ='1' then    
                            delta <= delta +1;
					end if;
				end if;
			end process;

tempv(0) <= v(0);
   veq0: for i in 1 to t generate
       tempv(i) <= tempv(i-1) or r(i);
   end generate; 
   vequal0 <= tempv(tempv'high); 			

v_register: process(clk)
			begin
				if rising_edge(clk) then
					if load = '1' then
						v <= (0=> '0', others => '0');

					elsif en_op_0 = '1' then
                        if dg_flag = '1' then
							v <= r;
							v_dr_shift_count <= unsigned(n_in) - unsigned(n_count);  
							count_v <= count_r;
						end if;
						
					elsif en_op_2 = '1' then	
						if vequal0 = '0' then
                           count_v <= (others => '0');
                        end if;

					end if;
				end if;
			end process;

tempr(0) <= r(0);
   req0: for i in 1 to t generate
       tempr(i) <= tempr(i-1) or r(i);
   end generate; 
   requal0 <= tempr(tempr'high); 
			
r_register: process(clk)
                 begin
                    if rising_edge(clk) then
                       if load = '1' then
                          r <= (0=> '1', others => '0');
                          
                        elsif en_op_0 = '1' then
                            if dg_flag = '1' then
                                 r <= signed(rv_adder_out(t downto 0));
                                count_r <= count_v; 
                            end if;
                            
                       elsif en_op_1 ='1' then
                            if  g_in_reg_2(0) = '1' then
                                r <= signed(rv_adder_out(t downto 0));                                  
                                if count_r >= count_v then
                                    count_r <= count_r;
                                else
                                    count_r <= count_v;
                                end if;
                            end if;
                                                                            
                       elsif en_op_2 = '1' then
                            if requal0 = '0' then
                               count_r <= (others => '0');
                            else
                               count_r <= count_r +1;
                            end if;
                       
                             
                       end if;     
                    end if;
                 end process;

 
counter_n: process(clk)
           begin
            if rising_edge(clk) then
                if init = '1' then
                    n_count <= n_in; 
                elsif en_n = '1' then
                    n_count <= std_logic_vector((unsigned(n_count) - 1));
                end if;    
            end if;
           end process;
 
counter_t: process(clk)
           begin
            if rising_edge(clk) then
                if init = '1' then
                    t_count <= t_in; 
                elsif en_t = '1' then
                    t_count <= std_logic_vector((unsigned(t_count) - 1));
                end if;    
            end if;
           end process;

 p_fsm: process(clk)
    begin
        if rising_edge(clk) then
            if (rst = '1') then   
            else
				if (state = S_RESET) then
					state           <= S_WAIT_START;
				elsif (state = S_WAIT_START) then
			         if start = '1' then
			             state <= S_MINUS_INIT;
			         end if;
				 
                elsif(state = S_MINUS_INIT) then
					state <= S_PROCESS; 
				    
				 elsif (state = S_PROCESS) then
				    if t_count =  one then
				        state <= S_DONE;
				    else
				        if done_add_rv = '1' then
    				        state <= S_STALL_1;
    				    end if;
				    end if;
             
                 elsif (state = S_STALL_1) then                    
                    state <= S_WAIT_ADD;
                 
                 elsif (state = S_WAIT_ADD) then
                    if done_add_1 = '1' then
                        state <= S_STALL_2;
                    else
                        state <= S_WAIT_ADD;
                    end if;
                    
                 elsif (state = S_STALL_2) then                   
                    state <= S_MINUS_INIT;
                    
                                   
				 elsif (state = S_DONE) then
				    state <= S_WAIT_START;
				    
				end if;
			   end if;

         end if;
    end process;
    
	done <= done_s; 
    p_comb: process(state, start, t_count, done_add_1, done_add_rv)
    begin
        --! Default values
        load <= '0';
        init <= '0';
        done_s <= '0';
        en_n <= '0';
        en_t <= '0';
        en_op_0 <= '0';
        en_op_1 <= '0';
        en_op_2 <= '0';    
        start_add <= '0';      
        minus <= '0';
		
        
		if (state = S_RESET) then

		elsif (state = S_WAIT_START) then
		  if start = '1' then
            load <= '1';
            init <= '1';
		  end if;
		
		elsif(state = S_MINUS_INIT) then
			start_add <= '1';
			minus <= '1';	
         	 
		elsif state = S_PROCESS then
		     if t_count /=  one then
		        if done_add_rv = '1' then
    		        en_op_0 <= '1'; 
    		    end if;   
			 end if;
					
        elsif (state =S_STALL_1) then
            start_add <= '1';
        
        elsif (state =S_WAIT_ADD) then
            if done_add_1 = '1' then
              en_op_1 <= '1';
            end if;      
        elsif (state = S_STALL_2) then
            en_op_2 <= '1';
            en_t <= '1';
            en_n <= '1';
                          
        elsif state = S_DONE then
         	done_s <= '1';
         				        
		end if;
		end process;	
		
end Behavioral;
