---------------------------------------------------------------------------------------
-- This file is a simple testbench for testing fast_gcd module with 434-bit input width 

-- Copyright (C) 2021
-- Authors: Sanjay Deshpande <sanjay.deshpande@yale.edu>
          
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software Foundation,
-- Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

----------------------------------------------------------------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.ALL;
use work.gcd_pkg.all; 

ENTITY fast_gcd_tb_434 IS  
END fast_gcd_tb_434;
 
ARCHITECTURE behavior OF fast_gcd_tb_434 IS 
 
    -- Component Declaration for the Unit Under Test (UUT)
 
    COMPONENT fast_gcd

  Port (
         clk          : in std_logic;
         rst        : in std_logic;
         start        : in std_logic;
         done         : out std_logic;
         f_in         : in STD_LOGIC_VECTOR (K-1 downto 0); 
         g_in         : in STD_LOGIC_VECTOR (K-1 downto 0);
         inv_out      : out STD_LOGIC_VECTOR(K-1 downto 0) 
         );
    END COMPONENT;
    

   --Inputs
   signal clk : std_logic := '0';
   signal rst : std_logic := '0';
   signal start : std_logic := '0';
   signal f_in : std_logic_vector(K-1 downto 0) := (others => '0');
   signal g_in : std_logic_vector(K-1 downto 0) := (others => '0');

 	--Outputs
   signal inv_out : std_logic_vector(K-1 downto 0);
   signal done : std_logic;
   
   -- Clock period definitions
   constant clk_period : time := 10 ns;
 
BEGIN
 
	-- Instantiate the Unit Under Test (UUT)
   uut: fast_gcd 
          PORT MAP (
          clk => clk,
          rst => rst,
          start => start,
          done => done,
          f_in => f_in,
          g_in => g_in,
          inv_out => inv_out
          );

   -- Clock process definitions
   clk_process :process
   begin
		clk <= '0';
		wait for clk_period/2;
		clk <= '1';
		wait for clk_period/2;
   end process;
 

   -- Stimulus process
   stim_proc: process
   begin		
      -- hold reset state for 100 ns.
      wait for 100 ns;	

      wait for 10*clk_period;

      -- insert stimulus here 
		
		start <= '1';
		
		f_in <= "11" & x"ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff49"; 
		g_in <= "11" & x"bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb11";

		wait for clk_period;
		
		start <= '0';
		
		-- expected output when done = 1 is inv_out =  0xf
	
      wait;
   end process;

END;
