# Fast constant-time GCD sequential implementation - Integer Case

This is a sequential hardware implementation of modular Inverse using Constant time GCD described in https://eprint.iacr.org/2019/266.pdf.

The design is generic and currently verified for primes (2^255 - 19),  (2^434 - 183), (2^1279 - 1), and  (2^2048 - 1557).

When integrated with the applications the design expects to use the modular multiplication and reduction unit from the application.

For choosing between 255, 434, 1279, 2048 - bit designs  uncomment the appropriate constants from the generic section of divstep2 code. The default is set to 255-bit design - 256-bit architecture.



- 'rtl' - consists of design files

- 'tb' - consists of testbenches for each case

  - divstep2_tb_255.vhd  - 255-bit design (set to 256-bit architecture, can be changed to other architectures by uncommenting required architecture) 

  - divstep2_tb_434.vhd - 434-bit design (set to 256-bit architecture, can be changed to other architectures by uncommenting required architecture)

  - divstep2_tb_1279.vhd - 1279-bit design (set to 256-bit architecture, can be changed to other architectures by uncommenting required architecture)  

  - divstep2_tb_2048.vhd - 2048-bit design (set to 256-bit architecture, can be changed to other architectures by uncommenting required architecture)

    ​	

    

### Fast GCD Algorithm 

```
# code derived from https://eprint.iacr.org/2019/266.pdf.
def iterations(d):
    if d < 46: 
        rounds = (49*d+80)//17  #// does floor division
    else: 
        rounds = (49*d+57)//17
    return rounds
    
def divsteps2(n,delta,f,g):
    v,r = 0,1
    m = n
    for n in range(m): 
        mask1 = (delta > 0) and (g&1 ==1)
        mask0 = (delta <= 0) or (g&1 ==0)
        delta,f,g,v,r = (mask0*delta + mask1*(-delta)), 
           (mask0*f + mask1*g), (mask0*g + mask1*(-f)), 
           (mask0*v + mask1*r), (mask0*r + mask1*(-v)) 
        g0 = g&1
        delta,g,r = 1+delta,(g+g0*f)/2,(r+g0*v)/2
        g = ZZ(g) #zz() means Integer Ring
    v_out = sign(f)*ZZ(v*2^(m-1))
    return v_out
    
def fast_gcd(f,g):
    d = max(f.nbits(),g.nbits())
    m = iterations(d)
    precomp = Integers(f)((f+1)/2)^(m-1)
    v_out = divsteps2(m,1,f,g)
    inverse = ZZ(v_out*precomp) 
    return inverse
```



This hardware design performs all operations from divstep2 module. When integrated with the applications the design expects to use the modular multiplication and reduction unit from the application to perform the final step ("inverse = ZZ(v_out*precomp)"). 
