----------------------------------------------------------------------------------
-- This file is sequential module for divstep2 function from fast GCD inverse for 
-- integers algorithm 

-- Copyright (C) 2021
-- Authors: Sanjay Deshpande <sanjay.deshpande@yale.edu>
          
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software Foundation,
-- Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

-- ramsize = t/d
-- ADDR_WIDTH = log2(ramsize)
-- K - fieldwidth + extra bits to make it divisible by 'd'
-- t = output of iterations function from Fast GCD algorithm + extra bits to make it 
-- 	   divisible by 'd' and result in 2 power number
-- d - number of divisions of 't' to make it a 2 power
-- x = d (should be even, sequential variable shifter setting)
-- y = d/2 (sequential variable shifter setting)
-- w = t
-- logw = log2(w)
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.NUMERIC_STD.ALL;
use IEEE.math_real.all;

entity divsteps2 is
    generic ( 

--========================================
-- 255-bit designs 
--======================================== 

        --==============================
        -- 256-bit architecture 
        --============================== 
  
		    ramsize : integer := 4;
		    ADDR_WIDTH : integer := 2;				
            K : integer := 256;
            d : integer := 4;
            logt : integer := 10;
            t : integer := 1024;
            x       : integer := 4;
            y       : integer := 2;
			w : integer := 1024;
			logw : integer := 10

        --==============================
        -- 128-bit architecture 
        --============================== 
			
--		    ramsize : integer := 6;
--		    ADDR_WIDTH : integer := 3;				
--            K : integer := 264;
--            d : integer := 6;
--            logt : integer := 10;
--            t : integer := 768;
--            x       : integer := 6;
--            y       : integer := 3;
--			w : integer := 768;
--			logw : integer := 10

        --==============================
        -- 64-bit architecture 
        --============================== 

--		    ramsize : integer := 12;
--		    ADDR_WIDTH : integer := 4;				
--            K : integer := 264;
--            d : integer := 12;
--            logt : integer := 10;
--            t : integer := 768;
--            x       : integer := 12;
--            y       : integer := 6;
--			w : integer := 768;
--			logw : integer := 10

        --==============================
        -- 32-bit architecture 
        --============================== 

--		    ramsize : integer := 24;
--		    ADDR_WIDTH : integer := 5;				
--            K : integer := 288;
--            d : integer := 24;
--            logt : integer := 10;
--            t : integer := 768;
--            x       : integer := 24;
--            y       : integer := 12;
--			w : integer := 768;
--			logw : integer := 10

        --==============================
        -- 16-bit architecture 
        --============================== 

--		    ramsize : integer := 48;
--		    ADDR_WIDTH : integer := 6;				
--            K : integer := 288;
--            d : integer := 48;
--            logt : integer := 10;
--            t : integer := 768;
--            x       : integer := 48;
--            y       : integer := 24;
--			w : integer := 768;
--			logw : integer := 10


--========================================
-- 434-bit designs 
--======================================== 

        --==============================
        -- 256-bit architecture 
        --============================== 
 
--             ramsize : integer := 6;
--		    ADDR_WIDTH : integer := 3;				
--            K : integer := 438;
--            d : integer := 6;
--            logt : integer := 11;
--            t : integer := 1536;
--            x       : integer := 6;
--            y       : integer := 3;
--			w : integer := 1536;
--			logw : integer := 11

        --==============================
        -- 128-bit architecture 
        --============================== 			
--		    ramsize : integer := 10;
--		    ADDR_WIDTH : integer := 4;				
--            K : integer := 440;
--            d : integer := 10;
--            logt : integer := 11;
--            t : integer := 1280;
--            x       : integer := 10;
--            y       : integer := 5;
--			w : integer := 1280;
--			logw : integer := 11

        --==============================
        -- 64-bit architecture 
        --============================== 
--		    ramsize : integer := 20;
--		    ADDR_WIDTH : integer := 5;				
--            K : integer := 440;
--            d : integer := 20;
--            logt : integer := 11;
--            t : integer := 1280;
--            x       : integer := 20;
--            y       : integer := 10;
--			w : integer := 1280;
--			logw : integer := 11

        --==============================
        -- 32-bit architecture 
        --============================== 			
--			ramsize : integer := 40;
--		    ADDR_WIDTH : integer := 6;				
--            K : integer := 440;
--            d : integer := 40;
--            logt : integer := 11;
--            t : integer := 1280;
--            x       : integer := 40;
--            y       : integer := 20;
--			w : integer := 1280;
--			logw : integer := 11

        --==============================
        -- 16-bit architecture 
        --==============================            
--            ramsize : integer := 80;
--		    ADDR_WIDTH : integer := 7;				
--            K : integer := 480;
--            d : integer := 80;
--            logt : integer := 11;
--            t : integer := 1280;
--            x       : integer := 80;
--            y       : integer := 40;
--			w : integer := 1280;
--			logw : integer := 11


--========================================
-- 1279-bit designs 
--======================================== 

        --==============================
        -- 256-bit architecture 
        --==============================  
-- 			ramsize : integer := 16;
--		    ADDR_WIDTH : integer := 5;				
--            K : integer := 1280;
--            d : integer := 16;
--            logt : integer := 12;
--            t : integer := 4096;
--            x       : integer := 16;
--            y       : integer := 8;
--			w : integer := 4096;
--			logw : integer := 12

        --==============================
        -- 128-bit architecture 
        --============================== 
	        
--			ramsize : integer := 30;
--		    ADDR_WIDTH : integer := 5;				
--            K : integer := 1290;
--            d : integer := 30;
--            logt : integer := 12;
--            t : integer := 3840;
--            x       : integer := 30;
--            y       : integer := 15;
--			w : integer := 3840;
--			logw : integer := 12		

        --==============================
        -- 64-bit architecture 
        --============================== 
--            ramsize : integer := 58;
--		    ADDR_WIDTH : integer := 6;				
--            K : integer := 1392;
--            d : integer := 58;
--            logt : integer := 12;
--            t : integer := 3712;
--            x       : integer := 58;
--            y       : integer := 29;
--			w : integer := 3712;
--			logw : integer := 12

        --==============================
        -- 32-bit architecture 
        --============================== 
--			ramsize : integer := 116;
--		    ADDR_WIDTH : integer := 7;				
--            K : integer := 1392;
--            d : integer := 116;
--            logt : integer := 12;
--            t : integer := 3712;
--            x       : integer := 116;
--            y       : integer := 58;
--			w : integer := 3712;
--			logw : integer := 12

        --==============================
        -- 16-bit architecture 
        --============================== 
--			ramsize : integer := 232;
--		    ADDR_WIDTH : integer := 8;				
--            K : integer := 1392;
--            d : integer := 232;
--            logt : integer := 12;
--            t : integer := 3712;
--            x       : integer := 232;
--            y       : integer := 116;
--			w : integer := 3712;
--			logw : integer := 12


 --========================================
-- 2048-bit designs 
--======================================== 

        --==============================
        -- 256-bit architecture 
        --============================== 
--			ramsize : integer := 24;
--		    ADDR_WIDTH : integer := 5;				
--            K : integer := 2064;
--            d : integer := 24;
--            logt : integer := 13;
--            t : integer := 6144;
--            x       : integer := 24;
--            y       : integer := 12;
--			w : integer := 6144;
--			logw : integer := 13

        --==============================
        -- 128-bit architecture 
        --============================== 
--		    ramsize : integer := 48;
--		    ADDR_WIDTH : integer := 6;				
--            K : integer := 2064;
--            d : integer := 48;
--            logt : integer := 13;
--            t : integer := 6144;
--            x       : integer := 48;
--            y       : integer := 24;
--			w : integer := 6144;
--			logw : integer := 13

        --==============================
        -- 64-bit architecture 
        --============================== 
--            ramsize : integer := 96;
--		    ADDR_WIDTH : integer := 7;				
--            K : integer := 2112;
--            d : integer := 96;
--            logt : integer := 13;
--            t : integer := 6144;
--            x       : integer := 96;
--            y       : integer := 48;
--			w : integer := 6144;
--			logw : integer := 13

        --==============================
        -- 32-bit architecture 
        --============================== 
--			ramsize : integer := 192;
--		    ADDR_WIDTH : integer := 8;				
--            K : integer := 2112;
--            d : integer := 192;
--            logt : integer := 13;
--            t : integer := 6144;
--            x       : integer := 192;
--            y       : integer := 96;
--			w : integer := 6144;
--			logw : integer := 13

        --==============================
        -- 16-bit architecture 
        --============================== 
--            ramsize : integer := 384;
--		    ADDR_WIDTH : integer := 9;				
--            K : integer := 2304;
--            d : integer := 384;
--            logt : integer := 13;
--            t : integer := 6144;
--            x       : integer := 384;
--            y       : integer := 192;
--			w : integer := 6144;
--			logw : integer := 13					  
    );    
    Port ( clk          : in STD_LOGIC;
           rst          : in STD_LOGIC;
           start        : in STD_LOGIC;
           valid        : out STD_LOGIC;
           done         : out STD_LOGIC;
           n_in         : in STD_LOGIC_VECTOR (logt-1 downto 0);
           t_in         : in STD_LOGIC_VECTOR (logt-1 downto 0);
           f_in         : in STD_LOGIC_VECTOR (K/d-1 downto 0);
           g_in         : in STD_LOGIC_VECTOR (K/d-1 downto 0);
           M2Q_v_out    : out STD_LOGIC_VECTOR (t/d-1 downto 0)
           
           );
end divsteps2;

architecture Behavioral of divsteps2 is
signal f_in_reg, g_in_reg : std_logic_vector(K/d - 1 downto 0);
signal f_in_reg_2, g_in_reg_2 : std_logic_vector(K/d - 1 downto 0);
signal notf : std_logic_vector(K/d - 1 downto 0);
signal notv : std_logic_vector(w/d - 1 downto 0);
signal n_count : std_logic_vector(logt-1 downto 0) := (others => '0');
signal t_count : std_logic_vector(logt-1 downto 0) := (others => '0');
constant one    : std_logic_vector(logt-1 downto 0) := (0 => '1', others => '0');
signal en_n, init, en_t : std_logic := '0';
type load_state is (S_RESET_L, S_WAIT_START_L, S_INIT_LOAD, S_TEST_ODD, S_LOAD_FIRST, S_LOAD_L, S_LOAD_DONE_L, S_SHIFT_L, S_ADD, S_STALL_0, S_STALL_1, S_STALL_2, S_STALL_3,S_STALL_4,S_V_RIGHT_SHIFT, S_MUL_WITH_M_1, S_STALL_5, S_FINAL_SHIFT_V, S_DONE);
signal state_load        			: load_state;
signal done_load :std_logic :='0';
--RAM Signals
signal wea : std_logic := '0';
signal web : std_logic := '0';
signal address_a : std_logic_vector(ADDR_WIDTH-1 downto 0) := (others => '0');
signal address_b : std_logic_vector(ADDR_WIDTH-1 downto 0) := (others => '0');
signal ram_out_a, ram_out_b : std_logic_vector(K/d-1 downto 0) := (others => '0');
signal count_chunks, count_chunks_reg : std_logic_vector(ADDR_WIDTH-1 downto 0) := (others => '0');
signal f_ram_out_a, f_ram_out_b : std_logic_vector(K/d-1 downto 0) := (others => '0');
signal f_ram_in_a, f_ram_in_b : std_logic_vector(K/d-1 downto 0) := (others => '0');
signal f_addr_a : std_logic_vector(ADDR_WIDTH-1 downto 0) := (others => '0');
signal f_addr_b : std_logic_vector(ADDR_WIDTH-1 downto 0) := (others => '0');
signal f_wea, f_web : std_logic := '0';
signal g_ram_out_a, g_ram_out_b : std_logic_vector(K/d-1 downto 0) := (others => '0');
signal g_ram_in_a, g_ram_in_b : std_logic_vector(K/d-1 downto 0) := (others => '0');
signal g_addr_a : std_logic_vector(ADDR_WIDTH-1 downto 0) := (others => '0');
signal g_addr_b : std_logic_vector(ADDR_WIDTH-1 downto 0) := (others => '0');
signal g_wea, g_web, g_wea_reg, g_web_reg : std_logic := '0';
signal shift_n_add_fg : std_logic := '0';
signal r_ram_out_a, r_ram_out_b : std_logic_vector(w/d-1 downto 0) := (others => '0');
signal r_ram_in_a, r_ram_in_b : std_logic_vector(w/d-1 downto 0) := (others => '0');
signal r_addr_a : std_logic_vector(ADDR_WIDTH-1 downto 0) := (others => '0');
signal r_addr_b : std_logic_vector(ADDR_WIDTH-1 downto 0) := (others => '0');
signal r_wea, r_web : std_logic := '0';
signal load_r_1 : std_logic := '0';
signal shift_n_add_r : std_logic := '0';
constant zero : std_logic_vector(t/d-1 downto 0) := (others=> '0');
signal v_input_to_shift, r_input_to_shift : std_logic_vector(w/d-1 downto 0) := (others => '0');
signal v_ram_out_a, v_ram_out_b : std_logic_vector(w/d-1 downto 0) := (others => '0');
signal v_ram_in_a, v_ram_in_b : std_logic_vector(w/d-1 downto 0) := (others => '0');
signal v_addr_a : std_logic_vector(ADDR_WIDTH-1 downto 0) := (others => '0');
signal v_addr_b : std_logic_vector(ADDR_WIDTH-1 downto 0) := (others => '0');
signal v_wea, v_web : std_logic := '0';
signal en_op_0, en_op_1, en_op_2 : std_logic := '0';
signal requal0, vequal0 : std_logic := '0';
signal tempr, tempv : std_logic_vector(w/d-1 downto 0) := (others => '0');
signal fplusg, fplusg_reg : std_logic_vector(K/d downto 0) := (others => '0');
signal carry_minus : std_logic := '0';
signal carry_fg : std_logic := '0';
signal carry_minus_en : std_logic := '0';
signal carry_fg_en : std_logic := '0';
signal init_fg : std_logic := '0';
signal init_minus : std_logic := '0';
signal fin_add_fg, fin_add_fg_reg, msb_g : std_logic := '0';
signal adder_in1 : std_logic_vector(K/d-1 downto 0) := (others => '0');
signal adder_in2 : std_logic_vector(K/d-1 downto 0) := (others => '0');
signal adder_carry_in : std_logic;
signal rv_adder_in1 : std_logic_vector(w/d-1 downto 0) := (others => '0');
signal rv_adder_in2 : std_logic_vector(w/d-1 downto 0) := (others => '0');
signal rv_adder_carry_in : std_logic := '0';
signal rv_carry_minus, carry_rv : std_logic := '0';
signal v_in_adder, r_in_adder : std_logic_vector(w/d-1 downto 0) := (others => '0');
signal rplusv, rplusv_reg : std_logic_vector(w/d downto 0) := (others => '0');
signal rv_carry_minus_en : std_logic := '0';
signal carry_rv_en : std_logic := '0';
signal init_rv : std_logic := '0';
signal test_odd_g : std_logic := '0';
signal odd_g, g_is_odd : std_logic := '0';
signal minus : std_logic := '0';
signal delta : std_logic_vector(logt downto 0) := (others => '0');
signal dg_flag : std_logic := '0';
signal minus_delta_en : std_logic := '0';
signal del_plus_1 : std_logic := '0';
signal count_v: std_logic_vector(logt-1 downto 0) := (1=> '0', others=> '0');
signal count_r: std_logic_vector(logt-1 downto 0) := (1 downto 0=> '0', others=> '0');
signal v_dr_shift_count: std_logic_vector(logt-1 downto 0) := (others=> '0');
signal shift_int : integer range 0 to t-1;
signal shift: std_logic_vector(logt-1 downto 0) := (others=> '0');
signal sh_ram_out_a, sh_ram_out_b : std_logic_vector(w/d-1 downto 0) := (others => '0');
signal sh_ram_in_a, sh_ram_in_b : std_logic_vector(w/d-1 downto 0) := (others => '0');
signal sh_addr_a : std_logic_vector(ADDR_WIDTH-1 downto 0) := (others => '0');
signal sh_addr_b : std_logic_vector(ADDR_WIDTH-1 downto 0) := (others => '0');
signal sh_wea, sh_web : std_logic := '0';
signal sh_valid, sh_valid_reg : std_logic := '0';
signal MUX_WORD: STD_LOGIC_VECTOR(w/y - 1 downto 0); 
signal OUTA_test: STD_LOGIC_VECTOR(w/y - 1 downto 0); 
signal shift_int_mask : integer range 0 to w/x;
signal shift_mul_chunks : integer range 0 to w;
signal shift_mul_chunks_plus_1 : integer range 0 to w;
signal mask: STD_LOGIC_VECTOR(w/x - 1 downto 0); 
signal shifted_output: STD_LOGIC_VECTOR(w/x - 1 downto 0); 
signal v_mul_2m_1 : std_logic := '0';
signal done_s: STD_LOGIC := '0'; 
signal final_valid: STD_LOGIC := '0'; 
signal sign_f : std_logic := '0';
signal v_final_shift : std_logic := '0';
signal minus_v_ram : std_logic_vector(w/d downto 0) := (others => '0');
signal m : STD_LOGIC_VECTOR (logt-1 downto 0);
signal final_valid_reg : STD_LOGIC := '0';
constant wby2 : integer := w/y;
constant wby4 : integer := w/x;
begin

m <= n_in - 1;
v_ram_in_a <= r_ram_out_b when minus = '1' and dg_flag = '1' else zero;

	mem_v: entity work.memram(memory) 
            generic map(
                            SIZE => ramsize,
                            ADDR_WIDTH => ADDR_WIDTH,
                            COL_WIDTH => t/d
                        )
            port map(
                        clk => clk,
                        wea => v_wea, 
                        web => v_web,  
                        address_a => v_addr_a,
                        address_b => v_addr_b,
                        datain_a => v_ram_in_a,
                        datain_b => v_ram_in_b,
                        dataout_a => v_ram_out_a,
                        dataout_b  => v_ram_out_b         
            );

r_ram_in_a <= rplusv(r_ram_in_a'high downto 0) when (minus = '1' and dg_flag = '1') or (shift_n_add_r = '1') else zero(r_ram_in_a'high downto 1) & load_r_1;

	mem_r: entity work.memram(memory) 
            generic map(
                            SIZE => ramsize,
                            ADDR_WIDTH => ADDR_WIDTH,
                            COL_WIDTH => t/d
                        )
            port map(
                        clk => clk,
                        wea => r_wea, 
                        web => r_web,  
                        address_a => r_addr_a,
                        address_b => r_addr_b,
                        datain_a => r_ram_in_a,
                        datain_b => r_ram_in_b,
                        dataout_a => r_ram_out_a,
                        dataout_b  => r_ram_out_b         
            );

notv <= not v_ram_out_b;
v_in_adder <= shifted_output when count_r > count_v else
	          v_ram_out_b;	                      
r_in_adder <= shifted_output when count_r < count_v and g_is_odd = '1' else r_ram_out_b;
--r_in_adder <= shifted_output when count_r < count_v  else r_ram_out_b;
rv_adder_in1 <= notv when minus = '1' else r_in_adder;
rv_adder_in2 <= (others => '0') when (minus = '1') or (g_is_odd = '0' and shift_n_add_r = '1' ) else v_in_adder;
rv_adder_carry_in <= rv_carry_minus when  minus = '1' else carry_rv;

adder_rv: entity work.adder(behavioral) 
            generic map(width => t/d)
            port map(
                       in1 => rv_adder_in1,       
                       in2 => rv_adder_in2,       
                       carry_in => rv_adder_carry_in,       
                       output => rplusv(rplusv'high-1 downto 0),
                       carry_out => rplusv(rplusv'high)        
            );

process(clk)
begin
	if rising_edge(clk) then
	   if init_minus = '1' then 
	       rv_carry_minus <= '1';
	   elsif rv_carry_minus_en = '1' then
		   rv_carry_minus <= rplusv(rplusv'high);
	   end if;
	end if;
end process;

process(clk)
begin
	if rising_edge(clk) then
	   if init_rv = '1' then 
	       carry_rv <= '0';
	   elsif carry_rv_en = '1' then
		   carry_rv <= rplusv(rplusv'high);
	   end if;
	end if;
end process;

process(clk)
begin
	if rising_edge(clk) then
		f_in_reg <= f_in;
		f_in_reg_2 <= f_in_reg;
end if;
end process;

f_ram_in_a <= g_ram_out_b when minus = '1' and dg_flag = '1' else f_in;	
	
mem_f: entity work.memram(memory) 
		generic map(
						SIZE => ramsize,
						ADDR_WIDTH => ADDR_WIDTH,
						COL_WIDTH => K/d
					)
		port map(
					clk => clk,
					wea => f_wea, 
					web => f_web,  
					address_a => f_addr_a,
					address_b => f_addr_b,
					datain_a => f_ram_in_a,
					datain_b => f_ram_in_b,
					dataout_a => f_ram_out_a,
					dataout_b  => f_ram_out_b         
		);

process(clk)
begin
	if rising_edge(clk) then
		g_in_reg <= g_in;
		g_in_reg_2 <= g_in_reg;
end if;
end process;

g_ram_in_a <=  fplusg(f_in'high downto 0) when minus = '1' and dg_flag = '1'  else 
			   msb_g & fplusg_reg(f_in'high downto 1) when shift_n_add_fg = '1' else 
			   g_in;	
msb_g <= 	fplusg(f_in'low) when fin_add_fg = '0' else fplusg_reg(fplusg'high-1);

	mem_g: entity work.memram(memory) 
            generic map(
                            SIZE => ramsize,
                            ADDR_WIDTH => ADDR_WIDTH,
                            COL_WIDTH => K/d
                        )
            port map(
                        clk => clk,
                        wea => g_wea, 
                        web => g_web,  
                        address_a => g_addr_a,
                        address_b => g_addr_b,
                        datain_a => g_ram_in_a,
                        datain_b => g_ram_in_b,
                        dataout_a => g_ram_out_a,
                        dataout_b  => g_ram_out_b         
            );
				

notf <= not f_ram_out_b;
adder_in1 <= (others => '0') when minus = '1' or (g_is_odd = '0' and shift_n_add_fg = '1' ) else f_ram_out_b;
adder_in2 <= notf when minus = '1' else g_ram_out_b;
adder_carry_in <= carry_minus when  minus = '1' else carry_fg;

adder_fg: entity work.adder(behavioral) 
            generic map(width => K/d)
            port map(
                       in1 => adder_in1,       
                       in2 => adder_in2,       
                       carry_in => adder_carry_in,       
                       output => fplusg(K/d - 1 downto 0),
                       carry_out => fplusg(fplusg'high)        
            );

process(clk)
begin
	if rising_edge(clk) then
        fplusg_reg <= fplusg;
        g_web_reg <= g_web;
        fin_add_fg_reg <= fin_add_fg;
    end if;
end process;

process(clk)
begin
	if rising_edge(clk) then
	   if init_minus = '1' then 
	       carry_minus <= '1';
	   elsif carry_minus_en = '1' then
		   carry_minus <= fplusg(fplusg'high);
	   end if;
	end if;
end process;

process(clk)
begin
	if rising_edge(clk) then
	   if init_fg = '1' then 
	       carry_fg <= '0';
	   elsif carry_fg_en = '1' then
		   carry_fg <= fplusg(fplusg'high);
	   end if;
	end if;
end process;

dg_flag <= '1' when signed(delta) > 0 and g_is_odd = '1' else '0';

--v_shift_count <= v_dr_shift_count;

                 
delta_register: process(clk)
			begin
				if rising_edge(clk) then
					if init = '1' then
						delta <= (0=> '1', others => '0');
					
					elsif minus_delta_en = '1' then
                        if dg_flag = '1' then
							delta <= 0-delta;
						end if;
					
					elsif del_plus_1 ='1' then    
                            delta <= delta +1;
					end if;
				end if;
			end process;

counter_n: process(clk)
           begin
            if rising_edge(clk) then
                if init = '1' then
                    n_count <= n_in; 
                elsif en_n = '1' then
                    n_count <= std_logic_vector((unsigned(n_count) - 1));
                end if;    
            end if;
           end process;
 
counter_t: process(clk)
           begin
            if rising_edge(clk) then
                if init = '1' then
                    t_count <= t_in; 
                elsif en_t = '1' then
                    t_count <= std_logic_vector((unsigned(t_count) - 1));
                end if;    
            end if;
           end process;
           
tempr(0) <= r_ram_in_a(0);
reqr: for i in 1 to r_ram_in_a'high generate
   tempr(i) <= tempr(i-1) or r_ram_in_a(i);
end generate; 

tempv(0) <= v_ram_out_b(0);
reqv: for i in 1 to v_ram_in_a'high generate
   tempv(i) <= tempv(i-1) or v_ram_in_a(i);
end generate; 
   
 p_fsm: process(clk)
    begin
        if rising_edge(clk) then
            if (rst = '1') then   
            else
				if (state_load = S_RESET_L) then
					state_load           <= S_WAIT_START_L;
					
				elsif (state_load = S_WAIT_START_L) then
			         if start = '1' then
			             state_load <= S_TEST_ODD;
			             f_addr_a <= (others => '0');
			             g_addr_a <= (others => '0');
			             f_addr_b <= (others => '0');
			             g_addr_b <= (others => '0');
			             
			             r_addr_a <= (others => '0');
			             r_addr_b <= (others => '0');
			             v_addr_a <= (others => '0');
			             v_addr_b <= (others => '0');
			             
			             --shifter signals
			             sh_addr_a <= (others => '0');
			             sh_addr_b <= (others => '0');
			             sh_valid <= '0';
			             count_chunks <= (others => '0');
			             sign_f <= '0';
			             
			         end if;
			         sh_valid <='0';
			         final_valid <='0';
			    
			    elsif (state_load = S_TEST_ODD) then
					state_load <= S_INIT_LOAD;
					odd_g <= g_in(0);
					f_addr_a <= f_addr_a + 1;
			        g_addr_a <= g_addr_a + 1;
			        
			        r_addr_a <= r_addr_a + 1;
			        v_addr_a <= v_addr_a + 1;
			        
			    elsif (state_load = S_INIT_LOAD) then
                     g_is_odd <= odd_g;
			         if (f_addr_a = ramsize -1) and (g_addr_a = ramsize -1) then                           
			             state_load <= S_LOAD_L;
			             f_addr_a <= (others => '0'); 
			             g_addr_a <= (others => '0'); 
			             r_addr_a <= (others => '0'); 
			             v_addr_a <= (others => '0'); 
			             
			             f_addr_b <= f_addr_b + 1;
			             g_addr_b <= g_addr_b + 1;
			             
			             r_addr_b <= r_addr_b + 1;
			             v_addr_b <= v_addr_b + 1;
			             
			              
			              
			         else
			             f_addr_a <= f_addr_a + 1;
			             g_addr_a <= g_addr_a + 1;
			             r_addr_a <= r_addr_a + 1;
			             v_addr_a <= v_addr_a + 1;
			         end if;
				
			         
			    elsif (state_load = S_LOAD_L) then
                    if n_count > 0 then
                         f_addr_a <= f_addr_a + 1;
                         g_addr_a <= g_addr_a + 1;
                         r_addr_a <= f_addr_a + 1;
                         v_addr_a <= g_addr_a + 1;
                         sh_addr_a <= sh_addr_a + 1;
                         
                         if g_addr_a = 0 then
                            if dg_flag = '1' then
                                odd_g <= g_ram_in_a(0);
                                test_odd_g <= g_ram_in_a(0); --test signal
                            end if;
                         end if;
                         
                         if (f_addr_b = ramsize -1) and (g_addr_b = ramsize -1) then
                             state_load <= S_STALL_0;
                             f_addr_b <= (others => '0'); 
                             g_addr_b <= (others => '0'); 
                             r_addr_b <= (others => '0'); 
                             v_addr_b <= (others => '0');
                             
                            
                             if dg_flag = '1' then
                                v_dr_shift_count <= (n_in) - (n_count);  
                             end if;
                         else
                             r_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned(r_addr_b + 1)) mod ramsize, ADDR_WIDTH));
						     v_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned(v_addr_b + 1)) mod ramsize, ADDR_WIDTH));
                             f_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned(f_addr_b + 1)) mod ramsize, ADDR_WIDTH));
						     g_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned(g_addr_b + 1)) mod ramsize, ADDR_WIDTH));
                         end if;
                     else
--                        state_load <=  S_DONE;
                        state_load <=  S_V_RIGHT_SHIFT;
                        sh_addr_a <= std_logic_vector(to_unsigned(to_integer(unsigned((ramsize - 1) - v_dr_shift_count(shift'high downto shift'high - ADDR_WIDTH + 1) + 1)) mod ramsize,ADDR_WIDTH));
			            sh_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned((ramsize - 1) - v_dr_shift_count(shift'high downto shift'high - ADDR_WIDTH + 1)) mod ramsize),ADDR_WIDTH));
			         
                     end if;
			     elsif (state_load = S_STALL_0) then
			         if dg_flag = '1' then
			              g_is_odd <= odd_g;
			         end if;
			         
					 test_odd_g <= '0'; --test signal
			         state_load <= S_SHIFT_L;
			         f_addr_a <= (others => '0');
			         g_addr_a <= (others => '0');
			         r_addr_a <= (others => '0');
			         v_addr_a <= (others => '0');
			         sh_addr_a <= std_logic_vector(to_unsigned(to_integer(unsigned((ramsize - 1) - shift(shift'high downto shift'high - ADDR_WIDTH + 1) + 1)) mod ramsize,ADDR_WIDTH));
			         sh_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned((ramsize - 1) - shift(shift'high downto shift'high - ADDR_WIDTH + 1)) mod ramsize),ADDR_WIDTH));
			      
                    if v_addr_b = 0 then --checking v for zero
                            vequal0 <= tempv(tempv'high);
                    end if;
			     
			     elsif (state_load <= S_SHIFT_L) then
			         if count_chunks = ramsize-1 then
			             sh_addr_a <= (others => '0');
			             sh_addr_b <= (others => '0');
						 r_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned(r_addr_b + 1)) mod ramsize, ADDR_WIDTH));
						 v_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned(v_addr_b + 1)) mod ramsize, ADDR_WIDTH));
						 r_addr_a <= r_addr_b;
						 f_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned(f_addr_b + 1)) mod ramsize, ADDR_WIDTH));
						 g_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned(g_addr_b + 1)) mod ramsize, ADDR_WIDTH));
						 g_addr_a <= r_addr_b;
			             count_chunks <= (others => '0');
			             state_load <= S_STALL_1;
			             sh_valid <= '1';
			             requal0 <= tempr(tempr'high) or requal0; --checking r value for zero
			             vequal0 <= tempv(tempv'high) or vequal0; --checking r value for zero
			             
			         else
	   	       	         sh_addr_a <= std_logic_vector(to_unsigned(to_integer(unsigned(sh_addr_a + 1)) mod ramsize, ADDR_WIDTH));
	   	       	         sh_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned(sh_addr_b + 1)) mod ramsize, ADDR_WIDTH));
	   	       	         count_chunks <= count_chunks +1;
	   	       	         sh_valid <='1';
						 if count_chunks > 0 then
							r_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned(r_addr_b + 1)) mod ramsize, ADDR_WIDTH));
						    v_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned(v_addr_b + 1)) mod ramsize, ADDR_WIDTH));
							r_addr_a <= r_addr_b;
							g_addr_a <= r_addr_b;
							
							if r_addr_a = 0 then
							   requal0 <= tempr(tempr'high); --checking r value for zero
							else
							   requal0 <= tempr(tempr'high) or requal0; --checking r value for zero
							end if;
							
							vequal0 <= tempv(tempv'high) or vequal0; --checking v value for zero
							
							if g_addr_a = 0 then
							     odd_g <= g_ram_in_a(0);
							     test_odd_g <= g_ram_in_a(0); --test signal
			                 end if;
							
						 end if;
						 f_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned(f_addr_b + 1)) mod ramsize, ADDR_WIDTH));
						 g_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned(g_addr_b + 1)) mod ramsize, ADDR_WIDTH));
	   	       	     end if;
				
				elsif (state_load = S_STALL_1) then
				    r_addr_b  <= (others => '0');
			        v_addr_b  <= (others => '0'); 
					state_load <= S_STALL_2;
					sh_valid <='0';
					r_addr_a <= r_addr_b;
					g_addr_a <= r_addr_b;
					requal0 <= tempr(tempr'high) or requal0; --checking r value for zero 
					vequal0 <= tempv(tempv'high) or vequal0; --checking r value for zero
					test_odd_g <= '0'; --test signal
			         
	   	       	elsif (state_load = S_STALL_2) then
					state_load <= S_STALL_3;
					r_addr_a <= r_addr_b;
					g_addr_a <= r_addr_b;
					requal0 <= tempr(tempr'high) or requal0; --checking r value for zero
					
					if  g_is_odd = '1' then
                        if count_r >= count_v then
                            count_r <= count_r;
                        else
                            count_r <= count_v;
                        end if;
                    end if;
				
				 elsif (state_load = S_STALL_3) then
				     state_load <= S_STALL_4;
					 if requal0 = '0' then
                        count_r <= (others => '0');
                     else
                        count_r <= count_r +1;
                     end if;
					 if vequal0 = '0' then
                            count_v <= (others => '0');
                     end if;
                     f_addr_a <= (others => '0');
                     g_addr_a <= (others => '0');
                     r_addr_a <= (others => '0');
                     v_addr_a <= (others => '0');
                     f_addr_b <= (others => '0');
                     g_addr_b <= (others => '0');
                     r_addr_b <= (others => '0');                     
                     v_addr_b <= (others => '0');
                     g_is_odd <= odd_g;
                     
                 
                 elsif (state_load = S_STALL_4) then
                     state_load <= S_LOAD_L;
                     if dg_flag = '1' then
                        v_dr_shift_count <= (n_in) - (n_count);  
                        count_v <= count_r;
                        count_r <= count_v;
                     end if;
                     f_addr_b <= f_addr_b + 1;
			         g_addr_b <= g_addr_b + 1;    
			         r_addr_b <= r_addr_b + 1;
			         v_addr_b <= v_addr_b + 1;
                 
                 elsif (state_load = S_V_RIGHT_SHIFT) then
                    if count_chunks = ramsize-1 then
			             sh_addr_a <= (others => '0');
			             sh_addr_b <= (others => '0');
			             v_addr_b  <= (others => '0');
			             f_addr_b  <= std_logic_vector(to_unsigned(ramsize - 1, ADDR_WIDTH));
			             state_load <= S_STALL_5;
--			             state_load <= S_DONE;
			             sh_valid <= '1';	
			             count_chunks <= (others=> '0');		             
			         else
	   	       	          sh_addr_a <= std_logic_vector(to_unsigned(to_integer(unsigned(sh_addr_a + 1)) mod ramsize, ADDR_WIDTH));
	   	       	          sh_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned(sh_addr_b + 1)) mod ramsize, ADDR_WIDTH));
	   	       	          count_chunks <= count_chunks +1;
	   	       	          sh_valid <='1';
	   	       	     end if;
	   	       	     
	   	       	  elsif (state_load = S_STALL_5) then
	   	       	       v_addr_b <= v_addr_b +1;
	   	       	       state_load <= S_MUL_WITH_M_1;
	   	       	       sign_f <= f_ram_out_b(f_ram_out_b'high);
	   	       	       sh_valid <= '0';
	   	       	  
	   	       	  elsif (state_load = S_MUL_WITH_M_1) then
                    if sh_addr_b = ramsize-1 then
                         count_chunks <= (others => '0');
			             sh_addr_a <= std_logic_vector(to_unsigned(to_integer(unsigned((ramsize - 1) - shift(shift'high downto shift'high - ADDR_WIDTH + 1) + 1)) mod ramsize,ADDR_WIDTH));
			             sh_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned((ramsize - 1) - shift(shift'high downto shift'high - ADDR_WIDTH + 1)) mod ramsize),ADDR_WIDTH));
			             state_load <= S_FINAL_SHIFT_V;

			         else
	   	       	          sh_addr_b <= sh_addr_b +1;
--	   	       	          v_addr_b <= v_addr_b +1;
	   	       	          count_chunks <= count_chunks +1;
	   	       	           if v_addr_b = ramsize -1 then
	   	       	               v_addr_b <= (others => '0');
	   	       	           else
	   	       	               v_addr_b <= v_addr_b +1;
	   	       	           end if;
	   	       	     end if;
	   	       	  
	   	       	     
	   	       	  elsif (state_load = S_FINAL_SHIFT_V) then
	   	       	       if count_chunks = ramsize-1 then
	   	       	         sh_addr_a <= (others => '0');
			             sh_addr_b <= (others => '0');
			             state_load <= S_DONE;
	   	       	       else
	   	       	          sh_addr_a <= std_logic_vector(to_unsigned(to_integer(unsigned(sh_addr_a + 1)) mod ramsize, ADDR_WIDTH));
	   	       	          sh_addr_b <= std_logic_vector(to_unsigned(to_integer(unsigned(sh_addr_b + 1)) mod ramsize, ADDR_WIDTH));
	   	       	          count_chunks <= count_chunks +1;
	   	       	          final_valid <='1';
	   	       	       end if;
	   	       	     
                  elsif (state_load = S_DONE) then
				    state_load <= S_WAIT_START_L;
				    final_valid <= '0';

				end if;
			   end if;
            count_chunks_reg <= count_chunks;
			sh_valid_reg <= sh_valid;
			valid <= final_valid;
			final_valid_reg <= final_valid; 
--			done <= done_s;
         end if;
    end process;
    
	
    p_comb: process(state_load, start, f_addr_a, f_addr_b, g_addr_a, g_addr_b, count_chunks_reg, count_chunks, n_count, dg_flag)
    begin
        --! Default value
        init <= '0';
        init_minus <= '0';
        init_fg <= '0';
        f_wea <= '0';
        f_web <= '0';
        g_wea <= '0';
        g_web <= '0';
        r_wea <= '0';
        r_web <= '0';
        v_wea <= '0';
        v_web <= '0';
        sh_wea <= '0';
        sh_web <= '0';
        load_r_1 <= '0';
        done_load <= '0';
        carry_minus_en <= '0';
        rv_carry_minus_en <= '0';
        carry_fg_en <= '0';
        minus <= '0';
        fin_add_fg <= '0';
        en_n <= '0';
        en_t <= '0';
        minus_delta_en <= '0';
        del_plus_1 <= '0';
        en_op_0 <= '0';
        en_op_1 <= '0';
        en_op_2 <= '0';
		carry_rv_en <= '0';
		init_rv <= '0';
		shift_n_add_r <= '0';
		shift_n_add_fg <= '0';
		v_final_shift <= '0';
		done_s <= '0';
        
      if (state_load = S_RESET_L) then

		elsif (state_load = S_WAIT_START_L) then
		     if start = '1' then
		        init_minus <= '1';
		        init <= '1';
		     end if;

		elsif (state_load = S_TEST_ODD) then
            f_wea <= '1';
		    g_wea <= '1';
		    r_wea <= '1';
		    v_wea <= '1';
		    load_r_1 <= '1';
		    		
		elsif (state_load = S_INIT_LOAD) then
	        f_wea <= '1';
		    g_wea <= '1';       
		    r_wea <= '1';       
		    v_wea <= '1';       

		  
		elsif (state_load = S_LOAD_L) then
		  if n_count > 0 then
              carry_minus_en <= '1';
              minus <= '1';
              sh_wea <= '1';
              rv_carry_minus_en <= '1';
              if dg_flag ='1' then
                  f_wea <= '1';
                  g_wea <= '1';
                  r_wea <= '1';
                  v_wea <= '1';		  
              end if;
               if (f_addr_b = ramsize -1) and (g_addr_b = ramsize -1) then
                  en_op_0 <= '1';
               end if;
		   end if;
		  

		
		elsif (state_load = S_STALL_0) then
		  minus <= '1';
		  sh_wea <= '1';
		  rv_carry_minus_en <= '1';
		  if dg_flag ='1' then
		      minus_delta_en <= '1';
              f_wea <= '1';
              g_wea <= '1';
              r_wea <= '1';
              v_wea <= '1';
           end if;		  
		  init_rv <= '1';
		  init_fg <= '1';

		
		elsif (state_load = S_SHIFT_L) then
			carry_fg_en <= '1';	
			if count_chunks_reg > 0 then
				carry_rv_en <= '1';
				r_wea <= '1';
				shift_n_add_r <= '1';
				
				g_wea <= '1';
							
			end if;
			if count_chunks > 0 then
				shift_n_add_fg <= '1';
			end if;
		
		elsif (state_load = S_STALL_1) then
			carry_rv_en <= '1';
			r_wea <= '1';
			shift_n_add_r <= '1';
			shift_n_add_fg <= '1';
			g_wea <= '1';
			
		elsif (state_load = S_STALL_2) then
			carry_rv_en <= '1';
			r_wea <= '1';
			shift_n_add_r <= '1';
			shift_n_add_fg <= '1';
			g_wea <= '1';
			fin_add_fg <= '1';
			
		elsif (state_load = S_STALL_3) then
		  en_n <= '1';
		  en_t <= '1';
		  del_plus_1 <= '1';
		  init_minus <= '1';
		  
        elsif (state_load = S_V_RIGHT_SHIFT) then
           if count_chunks = ramsize-1 then		
		      init_minus <= '1';
		   end if;          
		
		elsif (state_load = S_STALL_5) then
		  sh_web <= '1';
		  minus <= '1';
		
		elsif (state_load = S_MUL_WITH_M_1) then
		  sh_web <= '1';
		  rv_carry_minus_en <= '1';
		  minus <= '1';
		  	
        elsif (state_load = S_FINAL_SHIFT_V) then
          v_final_shift <= '1';
        
        elsif (state_load = S_DONE) then
		  done_s <= '1';
		  v_final_shift <= '1';	
		  	
		          
        end if;
		end process;	

--=========================variable shift logic=================
v_input_to_shift <= v_ram_in_a when dg_flag = '1' else v_ram_out_b;
r_input_to_shift <= r_ram_in_a when dg_flag = '1' else r_ram_out_b;
sh_ram_in_a <= v_input_to_shift when count_r > count_v else r_input_to_shift;
sh_ram_in_b <= v_ram_out_b when sign_f = '0' else rplusv(sh_ram_in_b'high downto 0);

mem_shift: entity work.memram(memory) 
            generic map(
                            SIZE => ramsize,
                            ADDR_WIDTH => ADDR_WIDTH,
                            COL_WIDTH => t/d
                        )
            port map(
                        clk => clk,
                        wea => sh_wea, 
                        web => sh_web,  
                        address_a => sh_addr_a,
                        address_b => sh_addr_b,
                        datain_a => sh_ram_in_a,
                        datain_b => sh_ram_in_b,
                        dataout_a => sh_ram_out_a,
                        dataout_b  => sh_ram_out_b         
            );

shift <= count_v - count_r when v_final_shift = '0' and signed(count_v - count_r) > 0 else 
         count_r - count_v when v_final_shift = '0' else
         m - v_dr_shift_count;        
MUX_WORD <= sh_ram_out_a & sh_ram_out_b;
shift_int <=  to_integer(unsigned(shift)) mod wby4;
u_shift_a: for position in 0 to wby2 - 1 generate
      OUTA_test(position) <= MUX_WORD((position - shift_int) mod (MUX_WORD'high + 1));-- when ((position  > shift_int) or (position = shift_int)) else '0'  ;
end generate u_shift_a;

u_shift_mask: for position in 0 to wby4 - 1 generate
      mask(position) <= '1' when ((position  > shift_int_mask) or (position = shift_int_mask)) else '0'  ;
end generate u_shift_mask;

shift_mul_chunks <=  wby4*to_integer(unsigned(count_chunks_reg));
shift_mul_chunks_plus_1 <=  wby4*(to_integer(unsigned(count_chunks_reg))+1) - 1;
shift_int_mask <=  to_integer(unsigned(shift)) mod wby4 when shift > shift_mul_chunks and shift <= shift_mul_chunks_plus_1 else
                   0 when shift <= shift_mul_chunks else 
                   wby4;
                   
process(clk)
begin  
    if rising_edge(clk) then                 
        shifted_output <= OUTA_test(w/y-1 downto w/x) and mask;
        done <= done_s;
    end if;
end process;

M2Q_v_out <= shifted_output;

end Behavioral;
