----------------------------------------------------------------------------------
-- This file is testbench for sequential module for divstep2 function for 255-bit 
--width from fast GCD inverse for integers algorithm

-- Copyright (C) 2021
-- Authors: Sanjay Deshpande <sanjay.deshpande@yale.edu>
          
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software Foundation,
-- Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
----------------------------------------------------------------------------------

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.ALL;
 
 
ENTITY divstep2_tb_255 IS
 generic ( 

--========================================
-- 255-bit designs 
--======================================== 

        --==============================
        -- 256-bit architecture 
        --============================== 
  
		    ramsize : integer := 4;
		    ADDR_WIDTH : integer := 2;				
            K : integer := 256;
            d : integer := 4;
            logt : integer := 10;
            t : integer := 1024;
            x       : integer := 4;
            y       : integer := 2;
			w : integer := 1024;
			logw : integer := 10

        --==============================
        -- 128-bit architecture 
        --============================== 
			
--		    ramsize : integer := 6;
--		    ADDR_WIDTH : integer := 3;				
--            K : integer := 264;
--            d : integer := 6;
--            logt : integer := 10;
--            t : integer := 768;
--            x       : integer := 6;
--            y       : integer := 3;
--			w : integer := 768;
--			logw : integer := 10

        --==============================
        -- 64-bit architecture 
        --============================== 

--		    ramsize : integer := 12;
--		    ADDR_WIDTH : integer := 4;				
--            K : integer := 264;
--            d : integer := 12;
--            logt : integer := 10;
--            t : integer := 768;
--            x       : integer := 12;
--            y       : integer := 6;
--			w : integer := 768;
--			logw : integer := 10

        --==============================
        -- 32-bit architecture 
        --============================== 

--		    ramsize : integer := 24;
--		    ADDR_WIDTH : integer := 5;				
--            K : integer := 288;
--            d : integer := 24;
--            logt : integer := 10;
--            t : integer := 768;
--            x       : integer := 24;
--            y       : integer := 12;
--			w : integer := 768;
--			logw : integer := 10

        --==============================
        -- 16-bit architecture 
        --============================== 

--		    ramsize : integer := 48;
--		    ADDR_WIDTH : integer := 6;				
--            K : integer := 288;
--            d : integer := 48;
--            logt : integer := 10;
--            t : integer := 768;
--            x       : integer := 48;
--            y       : integer := 24;
--			w : integer := 768;
--			logw : integer := 10
  
    );    
END divstep2_tb_255;
 
ARCHITECTURE behavior OF divstep2_tb_255 IS 
 
   -- Component Declaration for the Unit Under Test (UUT)
   --Inputs
   signal clk : std_logic := '0';
   signal rst : std_logic := '0';
   signal start : std_logic := '0';
   signal n_in : std_logic_vector(logt-1 downto 0) := (others => '0');
   signal t_in : std_logic_vector(logt-1 downto 0) := (others => '0');
   signal f_in : std_logic_vector(K/d-1 downto 0) := (others => '0');
   signal g_in : std_logic_vector(K/d-1 downto 0) := (others => '0');

 	--Outputs
   signal valid : std_logic;
   signal done : std_logic;
   signal M2Q_v_out : std_logic_vector(t/d-1 downto 0);

   -- Clock period definitions
   constant clk_period : time := 10 ns;
   signal f : std_logic_vector(K-1 downto 0) := (others => '0');
   signal g : std_logic_vector(K-1 downto 0) := (others => '0');
BEGIN
 
	-- Instantiate the Unit Under Test (UUT)
   uut: entity work.divsteps2(behavioral) 
	GENERIC MAP(
				ramsize => ramsize,
				ADDR_WIDTH => ADDR_WIDTH,				
            K => K,
            d => d,
            logt => logt,
            t => t,			  	
            x => x,			  	
            y => y,			  	
            w => w			  	
	)
	PORT MAP (
          clk => clk,
          rst => rst,
          start => start,
          valid => valid,
          done => done,
          n_in => n_in,
          t_in => t_in,
          f_in => f_in,
          g_in => g_in,
          M2Q_v_out => M2Q_v_out
        );

   -- Clock process definitions
   clk_process :process
   begin
		clk <= '0';
		wait for clk_period/2;
		clk <= '1';
		wait for clk_period/2;
   end process;
 

   -- Stimulus process
   stim_proc: process
   begin		
      -- hold reset state for 100 ns.
      wait for 100 ns;	

      wait for clk_period*10;

      -- insert stimulus here 
		start <= '1';
		n_in <=  "10" & x"e2";
		t_in <=  "10" & x"e3";
		wait for clk_period;
		start <= '0';
		
		for i in 1 to d loop
		  f_in <= f((i)*(k/d) - 1 downto (i-1)*(k/d)) ;
		  g_in <= g((i)*(k/d) - 1 downto (i-1)*(k/d)) ;
		  wait for clk_period;
		end loop;

		wait for clk_period;

      wait;
   end process;
   
f(256-1 downto 0) <= x"7fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffed"; --(2^255 - 19)
g(4-1 downto 0) <= x"f";

--Expexted output when valid == '1'  M2Q_v_out <= 0xffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff2eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeef1a000000000000000000000000000000000000000000000000000000000 (Its is a negative number represented in 2s compliment, few MSB 1s omitted)
--output is sent on port M2Q_v_out sequentially LSB to MSB 
 
END;
