----------------------------------------------------------------------------------

-- This file is a adder module used in fast_gcd inverse

-- Copyright (C) 2021
-- Authors: Sanjay Deshpande <sanjay.deshpande@yale.edu>
          
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software Foundation,
-- Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity adder is
    generic(
               width : integer := 256
    );
    Port ( in1 : in STD_LOGIC_VECTOR (width-1 downto 0);
           in2 : in STD_LOGIC_VECTOR (width-1 downto 0);
           carry_in : in STD_LOGIC;
           output : out STD_LOGIC_VECTOR (width-1 downto 0);
           carry_out : out STD_LOGIC);
--attribute use_dsp48 : string;
--attribute use_dsp48 of adder : entity is "yes";
end adder;

architecture Behavioral of adder is
signal output_temp : std_logic_vector(width downto 0);
begin

output_temp <= '0'&in1+in2+carry_in;
output <= output_temp(width-1 downto 0);
carry_out <= output_temp(width);
end Behavioral;
